# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2007 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://posterity.edgewall.org/wiki/License.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://posterity.edgewall.org/log/.

import email
from time import mktime

from paste.fileapp import DataApp
from paste.httpexceptions import HTTPNotFound
from paste.httpheaders import CONTENT_DISPOSITION

from posterity import app, log, request, session
from posterity.component import *
from posterity.core import IRoutesProvider, ISettingsPanel
from posterity.db import transaction
from posterity.msg import MessageRenderer
from posterity.model import *
from posterity.util.i18n import _
from posterity.util.mail import send_msg, reflow_quoted_text, EmailAddressList
from posterity.util.validators import *
from posterity.web import flash, redirect, send_template, validate


class LoginController(Component):

    implements(IRoutesProvider)

    def login(self, environ, start_response):
        if request.method == 'POST':
            password = encrypt_password(request.POST['password'])
            account = Account.get_by(username=request.POST['username'], 
                                     password=password)
            if account:
                log.info('User %s logged in' % account.username)
                session.account = account
                return redirect(start_response, '/tags/Inbox/convs')
        return send_template(start_response, 'login.html')    

    def logout(self, environ, start_response):
        if session.account:
            log.info('User %s logged out' % session.account.username)
        session.logout()
        return redirect(start_response, '/login')

    ## IRoutesProvider methods
    def add_routes(self, routes):
        routes.add('/login', GET=self.login, POST=self.login)
        routes.add('/logout', GET=self.logout)


class ConvController(Component):

    implements(IRoutesProvider)

    def perform_action(self, start_response, tag, do_action, sel, 
                      index_mode=False):
        """Implements common operations on a set of conversations.

        This method is used by many of the views controlled by this class.
        """
        redirect_url = '/tags/' + tag.name + '/convs'
        for s in sel:
            c = Conv.get(s)
            if do_action == 'archive':
                c.remove_tag('Inbox')
                c.mark_as_read(deleted=True)
            elif do_action == 'ham':
                c.remove_tag('Spam')
                c.undelete()
            elif do_action == 'spam':
                c.add_tag('Spam')
                c.mark_as_deleted()
                c.mark_as_read(deleted=True)
            elif do_action == 'delete':
                c.add_tag('Trash')
                c.mark_as_deleted()
                c.mark_as_read(deleted=True)
            elif do_action == 'undelete':
                c.remove_tag('Trash')
                c.undelete()
            elif do_action.startswith('tag_'):
                t = do_action[4:]
                c.add_tag(t)
                if not index_mode:
                    redirect_url = '/tags/%s/convs/%d' % (tag.name, c.id)
            elif do_action.startswith('untag_'):
                t = do_action[6:]
                c.remove_tag(t)
                if not index_mode:
                    redirect_url = '/tags/%s/convs/%d' % (tag.name, c.id)
        return redirect(start_response, redirect_url)

    def conv_index(self, environ, start_response):
        """The conversation index view.

        This view lists all conversations associated with a certain `tag`.
        """
        account = session.account
        tag = Tag.get_by(name=request.urlvars['tag'], account_id=account.id)
        if not tag:
            return HTTPNotFound()(environ, start_response)
        if request.method == 'POST':
            sel = request.POST.getall('sel')
            do_action = request.POST['do_action']
            return self.perform_action(start_response, tag, do_action, sel, 
                                       index_mode=True)
        offset = int(request.GET.get('offset', '0'))
        page_size = 25
        deleted = tag.name in ['Trash', 'Spam']
        convs = tag.get_convs(offset=offset, limit=page_size,
                              deleted=deleted)
        data = dict(convs=convs, tag=tag, offset=offset, page_size=page_size)
        return send_template(start_response, 'conv_index.html', data)

    def conv_show(self, environ, start_response):
        """Shows a specific conversation thread. 
    
        All messages associated with the conversation are listed below 
        each other in cronological order.
        """
        account = session.account
        tag = Tag.get_by(name=request.urlvars['tag'], account_id=account.id)
        conv = Conv.get(request.urlvars['conv'])
        if not conv or conv.account != account:
            return HTTPNotFound()(environ, start_response)
        if request.method == 'POST':
            do_action = request.POST['do_action']
            return self.perform_action(start_response, tag, do_action, [conv.id])
        # The 'Trash' and 'Spam' tags are special.
        if tag.name in ['Trash', 'Spam']:
            msgs = conv.deleted_msgs
        else:
            msgs = conv.msgs
        renderer = MessageRenderer()
        data = dict(tag=tag, conv=conv, msgs=msgs, renderer=renderer)
        return send_template(start_response, 'conv_show.html', data)

    ## IRoutesProvider methods
    def add_routes(self, routes):
        routes.add('/tags/{tag}/convs', GET=self.conv_index, 
                                        POST=self.conv_index)
        routes.add('/tags/{tag}/convs/{conv}', GET=self.conv_show, 
                                               POST=self.conv_show)


class MsgController(Component):

    implements(IRoutesProvider)

    @validate(schema=ComposerSchema, on_error='/msgs;compose_form')
    def create(self, environ, start_response):
        account = session.account
        send_msg(account, request.formvars['identity'],
                request.formvars['subject'], 
                reflow_quoted_text(request.formvars['body']),
                request.formvars['to'], request.formvars['cc'],
                request.formvars['in_reply_to'], 
                request.formvars['references'])
        flash('Message successfully sent')
        return redirect(start_response, '/tags/Sent%20Mail/convs')

    def compose_form(self, environ, start_response):
        account = session.account
        if not account.identities:
            flash('You need to add an identity before you can send an '
                      'email')
            return redirect(start_response, '/identities')
        data = dict(subject='', to='', cc='', body='',
                    in_reply_to='', references='')
        return send_template(start_response, 'composer.html', data)
    
    def reply_form(self, environ, start_response):
        return self._reply(environ, start_response, reply_all=False)
        
    def reply_all_form(self, environ, start_response):
        return self._reply(environ, start_response, reply_all=True)

    def _reply(self, environ, start_response, reply_all=False):
        account = session.account
        if not account.identities:
            flash('You need to add an identity before you can send an email')
            return redirect(start_response, '/identities')
        msg = Msg.get(request.urlvars['msg'])
        if not msg or msg.account != account:
            return HTTPNotFound()(environ, start_response)
        message = email.message_from_string(msg.payload)
        account_emails = [ident.email for ident in account.identities]
        account_emails = EmailAddressList(*account_emails)
        if 'Reply-To' in message:
            to = EmailAddressList(decode_header(message['Reply-To']))
        else:
            to = EmailAddressList(decode_header(message.get('From', '')))            
        if reply_all:
            orig_to = EmailAddressList(decode_header(message.get('To', '')))
            to = to + (orig_to - account_emails)
            cc = EmailAddressList(decode_header(message.get('Cc', '')))
            cc = cc - account_emails
        else:
            cc = ''
        subject = decode_header(message.get('Subject', ''))
        if subject and subject[:4].lower() != 're: ':
            subject = 'Re: ' + subject
        body = msg.quote_body()
        if 'Reply-To' in message:
            references = '%s \n\t%s' % (decode_header(message['References']),
                                        decode_header(message['Message-ID']))
        else:
            references = decode_header(message['Message-ID'])
            
        data = dict(body=body, to=unicode(to), cc=unicode(cc), subject=subject,
                    in_reply_to=msg.msg_id, references=references)
        return send_template(start_response, 'composer.html', data)

    def get_attachment(self, environ, start_response):
        account = session.account
        msg = Msg.get(request.urlvars['msg'])
        if not msg or msg.account != account:
            return HTTPNotFound()(environ, start_response)
        message = email.message_from_string(msg.payload)
        part_id = request.urlvars['id']
        for part in message.walk():
            content_id = part.get('Content-Id', '').strip('<>')
            filename = part.get_filename()
            if part_id in (content_id, filename):
                last_modified = mktime(msg.date.utctimetuple())
                content_type = part.get_content_type()
                app = DataApp('', content_type=content_type)
                app.set_content(part.get_payload(decode=True),
                                last_modified=last_modified)
                CONTENT_DISPOSITION.update(app.headers, attachment=True,
                                           filename=filename)
                return app(environ, start_response)
        return HTTPNotFound()(environ, start_response)

    ## IRoutesProvider methods
    def add_routes(self, routes):
        routes.add('/msgs', POST=self.create)
        routes.add('/msgs;compose_form', GET=self.compose_form)
        routes.add('/msgs/{msg};reply_form', GET=self.reply_form)
        routes.add('/msgs/{msg};reply_all_form', GET=self.reply_all_form)
        old_parser = routes.parser
        routes.parser = lambda x: x  
        routes.add(r'^/msgs/(?P<msg>\w+)/attachments/(?P<id>.+)$', 
                   GET=self.get_attachment)
        routes.parser = old_parser


class TagController(Component):

    implements(IRoutesProvider, ISettingsPanel)

    def delete(self, environ, start_response):
        """Handles DELETE requests for /tags/{tag}"""
        account = session.account
        tag = Tag.get_by(name=request.urlvars['tag'], account_id=account.id)
        if not tag or tag.account != account:
            return HTTPNotFound()(environ, start_response)
        assert tag.is_special == False
        log.info('Tag %s deleted by %s' % (tag.name, account.username))
        objectstore.delete(tag)
        flash('Tag deleted')
        return redirect(start_response, '/tags')

    @validate(schema=TagSchema, on_error='/tags')
    def create(self, environ, start_response):
        account = session.account
        tag = Tag(account=account, name=request.formvars['name'])
        log.info('Tag %s added by %s' % (tag.name, account.username))
        flash('Tag added')
        return redirect(start_response, '/tags')
            
    @validate(schema=TagSchema, on_error='/tags/{tag};edit_form')
    def update(self, environ, start_response):
        account = session.account
        tag = Tag.get_by(account=account, name=request.urlvars['tag'])
        if not tag or tag.account != account:
            return HTTPNotFound()(environ, start_response)
        assert tag.is_special == False
        tag.name = request.formvars['name']
        flash('Tag updated')
        return redirect(start_response, '/tags')

    def edit_form(self, environ, start_response):
        account = session.account
        tag = Tag.get_by(account=account, name=request.urlvars['tag'])
        if not tag or tag.account != account:
            return HTTPNotFound()(environ, start_response)
        data = dict(tag=tag)
        return send_template(start_response, 'tag_edit.html', data)

    def manage_form(self, environ, start_response):
        account = session.account
        conn = transaction.connection(Conv)
        s = select([conv_to_tag.c.tag_id, func.count(conv_to_tag.c.conv_date),
                    func.count(conv_to_tag.c.conv_date2)],
                   and_(Tag.c.account_id==account.id,
                        Tag.c.id==conv_to_tag.c.tag_id),
                   group_by=conv_to_tag.c.tag_id)
        conv_count = {}
        for tag_id, count_normal, count_deleted in conn.execute(s):
            conv_count[tag_id] = count_normal
        data = dict(account=account, conv_count=conv_count)
        return send_template(start_response, 'tags.html', data)

    ## IRoutesProvider methods
    def add_routes(self, routes):
        routes.add('/tags', GET=self.manage_form, POST=self.create)
        routes.add('/tags/{tag}', GET=self.edit_form,
                                  PUT=self.update, DELETE=self.delete)

    ## ISettingsPanel methods
    def get_settings_panels(self):
        yield dict(id='tags', label=_('Tags'), href='/tags')


class IdentityController(Component):
    
    implements(IRoutesProvider, ISettingsPanel)
    
    def manage_form(self, environ, start_response):
        return send_template(start_response, 'identity_list.html')

    def edit_form(self, environ, start_response):
        account = session.account
        ident = Identity.get(request.urlvars['ident'])
        if not ident or ident.account != account:
            return HTTPNotFound()(environ, start_response)
        data = dict(ident=ident)
        return send_template(start_response, 'identity_edit.html', data)

    @validate(schema=IdentitySchema, on_error='/identities')
    def create(self, environ, start_response):
        account = session.account
        ident = Identity(account=account, **request.formvars)
        flash('Identity added')
        log.info('Identity "%s" added by %s' % (ident.name, account.username))
        return redirect(start_response, '/identities')
            
    @validate(schema=IdentitySchema, on_error='/identities/{ident}')
    def update(self, environ, start_response):
        account = session.account
        ident = Identity.get(request.urlvars['ident'])
        if not ident or ident.account != account:
            return HTTPNotFound()(environ, start_response)
        ident.name = request.formvars['name']
        ident.email = request.formvars['email']
        ident.reply_to = request.formvars['reply_to']
        ident.organization = request.formvars['organization']
        ident.signature = request.formvars['signature']
        flash('Identity updated')
        return redirect(start_response, '/identities')

    def delete(self, environ, start_response):
        account = session.account
        ident = Identity.get(request.urlvars['ident'])
        if not ident or ident.account != account:
            return HTTPNotFound()(environ, start_response)
        log.info('Identity "%s" deleted by %s' % (ident.name, account.username))
        objectstore.delete(ident)
        flash('Identity deleted')
        return redirect(start_response, '/identities')

    ## IRoutesProvider methods
    def add_routes(self, routes):
        routes.add('/identities/{ident}', GET=self.edit_form, 
                                          PUT=self.update, 
                                          DELETE=self.delete)
        routes.add('/identities', GET=self.manage_form, POST=self.create)

    ## ISettingsPanel methods
    def get_settings_panels(self):
        yield dict(id='idents', label=_('Identities'), href='/identities')
