# -*- coding: utf-8 -*-
#
# Copyright (C) 2006-2007 Edgewall Software
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at http://posterity.edgewall.org/wiki/License.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://posterity.edgewall.org/log/.

from elixir import objectstore
from paste.httpexceptions import HTTPNotFound

from posterity import log, request, session
from posterity.component import Component, implements, Interface
from posterity.core import IRoutesProvider, ISettingsPanel, IMessageFilter
from posterity.model import Filter
from posterity.util.i18n import _
from posterity.util.validators import FilterSchema
from posterity.web import flash, redirect, send_template, validate


class FilterManager(Component):

    implements(IMessageFilter)

    ## IMessageFilter methods
    def incoming_message(self, account, msg):
        for rule in account.filters:
            if rule.matches(msg):
                rule.apply(msg)


class FilterController(Component):

    implements(IRoutesProvider, ISettingsPanel)

    def index(self, environ, start_response):
        return send_template(start_response, 'filter_list.html')

    @validate(schema=FilterSchema, on_error='/filters')
    def create(self, environ, start_response):
        filter = Filter(account=session.account,
                        **request.formvars)
        filter.enabled = True
        objectstore.flush()
        flash('Filter added')
        return redirect(start_response, '/filters')

    @validate(schema=FilterSchema)
    def update(self, environ, start_response):
        filter = Filter.get_by(id=request.urlvars['id'],
                               account=session.account)
        if not filter:
            return HTTPNotFound()(environ, start_response)
        for k, v in request.formvars.items():
            setattr(filter, k, v)
        objectstore.flush()
        flash('Filter updated')
        return redirect(start_response, '/filters')
    
    def delete(self, environ, start_response):
        filter = Filter.get_by(id=request.urlvars['id'],
                               account=session.account)
        if not filter:
            return HTTPNotFound()(environ, start_response)
        objectstore.delete(filter)
        flash('Filter deleted')
        return redirect(start_response, '/filters')

    def edit_form(self, environ, start_response):
        filter = Filter.get_by(id=request.urlvars['id'],
                               account=session.account)
        if not filter:
            return HTTPNotFound()(environ, start_response)
        data = dict(filter=filter)
        return send_template(start_response, 'filter_edit.html', data)

    ## IRoutesProvider methods
    def add_routes(self, routes):
        routes.add('/filters/{id}', GET=self.edit_form,
                                    PUT=self.update,
                                    DELETE=self.delete)
        routes.add('/filters', GET=self.index, POST=self.create)

    ## ISettingsPanel methods
    def get_settings_panels(self):
        yield dict(id='filters', label=_('Filters'), href='/filters')
